/*++

Copyright (c) 1993  Digital Equipment Corporation

Module Name:

    perfsup.c

Abstract:

    This module implements support functions for the Alpha AXP 21064
    performance counter driver.

Author:

    Joe Notarangelo  26-Jun-1993

Environment:

    Kernel mode only.

Notes:


Revision History:

    Brad Waters       1-Sep-1994
    Added include of ints.h to defined uint64 which was added to the
    DEVICE_EXTENSION structure in perfev4.h.  Also added calls to trace
    facility.
--*/
//#include "ints.h"
#include "ntddk.h"
#include "dda21064.h"

#include "ipr_kernel_int_defs.h"

#include "ipr_driver_nt.h"

#include "trpfrm.h"
#include "perfev4.h"

#ifdef   TRCI
#include "TraceDrv.h"
#endif


NTSTATUS
EnablePerformanceCounter(
    PDEVICE_EXTENSION DeviceExtension
    );

VOID
DisablePerformanceCounter(
    PDEVICE_EXTENSION DeviceExtension
    );


BOOLEAN
ValidateEventCount(
    PAXP21064_SETCOUNTER EventCountParameters,
    AXP21064_PCCOUNTER PerformanceCounter
    )
/*++

Routine Description:

    This routine validates the parameters for a set event count ioctl
    are valid.

Arguments:

    EventCountParameters - Supplies a pointer to the event count parameters
                               to be validated.

    PerformanceCounter - Supplies the performance counter for which the
                             parameters are to be validated.

Return Value:

    TRUE is returned if the parameters are valid, otherwise false is
    returned.

--*/
{

#ifdef TRCD
    TrcPrint("ValidateEventCount, counter %d, freq %d\n",
        PerformanceCounter,EventCountParameters->EventCount);
#endif
    //
    // Validate the event count values as appropriate for the
    // given performance counter.
    //

    switch( PerformanceCounter ){

    //
    // Handle Performance Counter 0
    //

    case PerformanceCounter0:

        switch( EventCountParameters->EventCount ){

        //
        // The valid cases.
        //

        case CountEvents2xx12:
        case CountEvents2xx16:

            return TRUE;

        //
        // Otherwise, not valid.
        //

        default:

            return FALSE;

        } //end switch EventCountParameters->EventCount

    //
    // Handle Performance Counter 1
    //

    case PerformanceCounter1:

        switch( EventCountParameters->EventCount ){

        //
        // The valid cases.
        //

        case CountEvents2xx8:
        case CountEvents2xx12:

            return TRUE;

        //
        // Otherwise, not valid.
        //

        default:

            return FALSE;

        } //end switch EventCountParameters->EventCount

    //
    // Otherwise, invalid performance counter
    //

    default:

        return FALSE;

    } //end switch PerformanceCounter

}


//
// Define structure for passing parameters to the synchronized set
// routine.
//

typedef struct _EVENTCOUNT_SYNCH{
    PDEVICE_EXTENSION DeviceExtension;
    PAXP21064_SETCOUNTER EventCountParameters;
    NTSTATUS Status;
} EVENTCOUNT_SYNCH, *PEVENTCOUNT_SYNCH;

VOID
SetEventCountSynchronized(
    PEVENTCOUNT_SYNCH SynchronizedParameters
    );

NTSTATUS
SetEventCount(
    PDEVICE_EXTENSION DeviceExtension,
    PAXP21064_SETCOUNTER EventCountParameters
    )
/*++

Routine Description:

    This routine sets a new event count for the performance counter
    described by the specified device extension.

Arguments:

    DeviceExtension - Supplies a pointer to the device extension for the
                          performance counter to be modified.

    EventCountParameters - Supplies a pointer to the new event count
                               parameters for the performance counter.

Return Value:

    The status from the synchronized update routine is returned.

--*/
{
//    EVENTCOUNT_SYNCH SynchronizedParameters;

#ifdef DBG5
    DbgPrint("Enter SetEventCount\n\n");
    DbgBreakPoint();
#endif
/*
    SynchronizedParameters.DeviceExtension = DeviceExtension;
    SynchronizedParameters.EventCountParameters = EventCountParameters;

    KeSynchronizeExecution( DeviceExtension->InterruptObject,
                            (PKSYNCHRONIZE_ROUTINE)SetEventCountSynchronized,
                            &SynchronizedParameters );

    return SynchronizedParameters.Status;
    //
    // Set the new event count in the device extension.
    //

*/
    DeviceExtension->EventCount = EventCountParameters->EventCount;
    return STATUS_SUCCESS;
}

VOID
SetEventCountSynchronized(
    PEVENTCOUNT_SYNCH SynchronizedParameters
    )
/*++

Routine Description:

    This routine does the actual work of setting new event count parameters
    for a performance counter.  It is synchronized with the interrupt
    routine and all other writers of the device extension.

Arguments:

    SynchronizedParameters - Supplies the data used to set a new event count
                                 for a performance counter.

Return Value:

    None.  However, status is returned via the SynchronizedParameters
    structure.

--*/
{
    PDEVICE_EXTENSION deviceExtension;
    PAXP21064_SETCOUNTER EventCountParameters;
//    EVENT_COUNT EventCount;
    NTSTATUS Status;

    deviceExtension = SynchronizedParameters->DeviceExtension;
    EventCountParameters = SynchronizedParameters->EventCountParameters;

    //
    // Set the new event count in the device extension.
    //

    deviceExtension->EventCount = EventCountParameters->EventCount;

    //
    // If the counter is current started, then enable the counter to
    // begin collecting data.
    //

    Status = EnablePerformanceCounter( deviceExtension );

    SynchronizedParameters->Status = Status;

    return;

}


BOOLEAN
ValidateMuxControl(
    PAXP21064_SETMUXCONTROL MuxControlParameters,
    AXP21064_PCCOUNTER PerformanceCounter
    )
/*++

Routine Description:

    This routine validates the parameters for a set mux control ioctl
    are valid.

Arguments:

    MuxControlParameters - Supplies a pointer to the mux control parameters
                               to be validated.

    PerformanceCounter - Supplies the performance counter for which the
                             the mux control parameters must be validated.
Return Value:

    TRUE is returned if the parameters are valid, otherwise false is
    returned.

--*/
{

    //
    // Validate the event count values as appropriate for the
    // given performance counter.
    //

    switch( PerformanceCounter ){

    //
    // Handle Performance Counter 0
    //

    case PerformanceCounter0:

        switch( MuxControlParameters->MuxControl ){

        //
        // The valid cases.
        //

        case TotalIssues:
        case PipelineDry:
        case LoadInstruction:
        case PipelineFrozen:
        case BranchInstructions:
        case PalMode:
        case TotalCycles:
        case TotalNonIssues:
        case ExternalCounter0:

            return TRUE;

        //
        // Otherwise, not valid.
        //

        default:

            return FALSE;

        } //end switch MuxControlParameters->MuxControl

    //
    // Handle Performance Counter 1
    //

    case PerformanceCounter1:

        switch( MuxControlParameters->MuxControl ){

        //
        // The valid cases.
        //

        case DcacheMiss:
        case IcacheMiss:
        case DualIssues:
        case BranchMispredicts:
        case FPInstructions:
        case IntegerOperate:
        case StoreInstructions:
        case ExternalCounter1:

            return TRUE;

        //
        // Otherwise, not valid.
        //

        default:

            return FALSE;

        } //end switch MuxControlParameters->MuxControl

    //
    // Otherwise, invalid performance counter
    //

    default:

        return FALSE;

    } //end switch PerformanceCounter

}


//
// Define structure for passing parameters to the synchronized set
// routine.
//

typedef struct _MUXCONTROL_SYNCH{
    PDEVICE_EXTENSION DeviceExtension;
    PAXP21064_SETMUXCONTROL MuxControlParameters;
    NTSTATUS Status;
} MUXCONTROL_SYNCH, *PMUXCONTROL_SYNCH;

VOID
SetMuxControlSynchronized(
    PMUXCONTROL_SYNCH SynchronizedParameters
    );

NTSTATUS
SetMuxControl(
    PDEVICE_EXTENSION DeviceExtension,
    PAXP21064_SETMUXCONTROL MuxControlParameters
    )
/*++

Routine Description:

    This routine sets a new mux control for the performance counter
    described by the specified device extension.

Arguments:

    DeviceExtension - Supplies a pointer to the device extension for the
                          performance counter to be modified.

    MuxControlParameters - Supplies a pointer to the new mux control
                               parameters for the performance counter.

Return Value:

    The status of the synchronized update routine is returned.

--*/
{
/*
    MUXCONTROL_SYNCH SynchronizedParameters;

    SynchronizedParameters.DeviceExtension = DeviceExtension;
    SynchronizedParameters.MuxControlParameters = MuxControlParameters;

#ifdef DBGPM
    DbgPrint("SetMuxControl calling ...synch, DevExt %x, IntObj %x, SynchParm.DevExt %x, SynchParm.MuxCntl %x\n",
       DeviceExtension,DeviceExtension->InterruptObject,SynchronizedParameters.DeviceExtension,
       SynchronizedParameters.MuxControlParameters);
    DbgPrint("DevExt.counter %d, DevExt.connect %d, DevExt.DevObj %x\n",
       DeviceExtension->Counter,DeviceExtension->connected,DeviceExtension->DeviceObject);
    DbgBreakPoint();
#endif

    KeSynchronizeExecution( DeviceExtension->InterruptObject,
                            (PKSYNCHRONIZE_ROUTINE)SetMuxControlSynchronized,
                            &SynchronizedParameters );
#ifdef DBGPM
    DbgPrint("SetMuxControl returning from calling ...synch, status = %x\n",
       SynchronizedParameters.Status);
    DbgBreakPoint();
#endif
*/
    //
    // Set the new mux control in the device extension.
    //

    DeviceExtension->MuxControl = MuxControlParameters->MuxControl;

    return STATUS_SUCCESS;

}

VOID
SetMuxControlSynchronized(
    PMUXCONTROL_SYNCH SynchronizedParameters
    )
/*++

Routine Description:

    This routine does the actual work of setting new mux control parameters
    for a performance counter.  It is synchronized with the interrupt
    routine and all other writers of the device extension.

Arguments:

    SynchronizedParameters - Supplies the data used to set a new mux control
                                 for a performance counter.

Return Value:

    None.  However, the status of the operation is returned via the
    SynchronizedParameters structure.

--*/
{
    PDEVICE_EXTENSION deviceExtension;
    PAXP21064_SETMUXCONTROL MuxControlParameters;
//    EVENT_COUNT EventCount;
    NTSTATUS Status;

#ifdef DBGPM
    DbgPrint("SetMuxControlSynchronized enter, Synchronized parm %x\n",
       SynchronizedParameters);
    DbgBreakPoint();
#endif

    deviceExtension = SynchronizedParameters->DeviceExtension;
    MuxControlParameters = SynchronizedParameters->MuxControlParameters;

    //
    // Set the new mux control in the device extension.
    //

    deviceExtension->MuxControl = MuxControlParameters->MuxControl;

    //
    // If the counter is current started, then enable the counter to
    // begin collecting data.
    //

#ifdef DBGPM
    DbgPrint("SetMuxControlSynchronized calling EnablePerformanceCounter\n");
    DbgBreakPoint();
#endif

    Status = EnablePerformanceCounter( deviceExtension );

#ifdef DBGPM
    DbgPrint("SetMuxControlSynchronized return from EnablePerformanceCounter\n");
    DbgBreakPoint();
#endif

    SynchronizedParameters->Status = Status;

    return;

}


//
// Define structure for passing parameters to the synchronized set
// routine.
//

typedef struct _STARTCOUNTER_SYNCH{
    PDEVICE_EXTENSION DeviceExtension;
    NTSTATUS Status;
} STARTCOUNTER_SYNCH, *PSTARTCOUNTER_SYNCH;

VOID
StartCounterSynchronized(
    PSTARTCOUNTER_SYNCH SynchronizedParameters
    );

NTSTATUS
StartCounter(
    PDEVICE_EXTENSION DeviceExtension
    )
/*++

Routine Description:

    Start the counter for the performance counter described by the
    specified device extension.

Arguments:

    DeviceExtension - Supplies a pointer to the device extension for the
                          performance counter to be started.

Return Value:

    The status of the operation is returned.

--*/
{
    NTSTATUS Status;
/*
    STARTCOUNTER_SYNCH SynchronizedParameters;

    SynchronizedParameters.DeviceExtension = DeviceExtension;

    //
    // Synchronize the start routine.
    //

    KeSynchronizeExecution( DeviceExtension->InterruptObject,
                            (PKSYNCHRONIZE_ROUTINE)StartCounterSynchronized,
                            &SynchronizedParameters );

    if( !NT_SUCCESS(SynchronizedParameters.Status) ){
        return SynchronizedParameters.Status;
    }

    //
    // If the interrupt is not currently connected, then reconnect
    // it.
    //
*/
#ifdef DBGCRSH
    DbgPrint("Perfsup: calling EnablePerformanceCounter, DevExt %x\n",
       DeviceExtension);
#endif
    Status = EnablePerformanceCounter( DeviceExtension );
    if( !NT_SUCCESS(Status) ){
        return Status;
    }

#ifdef DBGCRSH
    DbgPrint("Perfsup: return from EnablePerformanceCounter, Status %x\n",
       Status);
#endif
    if( DeviceExtension->Started != TRUE ){

        Status = IoConnectInterrupt(
                     &DeviceExtension->InterruptObject,
//davew              (PKSERVICE_ROUTINE)PerformanceCounterInterrupt,
                     (PKSERVICE_ROUTINE)PCInterruptAddFrame,
                     (PVOID)DeviceExtension,
                     DeviceExtension->ISRSpin,
                     DeviceExtension->Vector,
                     DeviceExtension->Irql,
                     DeviceExtension->Irql,
                     Latched,
                     TRUE,
                     DeviceExtension->Affinity,
                     FALSE );

#ifdef DBGCRSH
       DbgPrint("Perfsup: return from IoConnectInterrupt, DevExt %x, status %x\n",
          DeviceExtension,Status);
#endif
        //
        // Then signify the operation was not successful.
        //

        if( Status != STATUS_SUCCESS ){
            AxpPrint((1, "Error reconnecting interrupt Status = %x\n",
                         Status ));
            return Status;
        }

    }

    //
    // Indicate that the performance counter has been successfully
    // started.
    //

    DeviceExtension->Started = TRUE;

    return Status;

}

VOID
StartCounterSynchronized(
    PSTARTCOUNTER_SYNCH SynchronizedParameters
    )
/*++

Routine Description:

    Start the counter described by the specified device extension.

Arguments:

    SynchronizedParameters - Supplies a pointer to the synchronized
                                 parameters for the start operation.

Return Value:

    None.  However, the a status is returned via the SynchronizedParameters.

--*/
{
    PDEVICE_EXTENSION deviceExtension;
    NTSTATUS Status;

    deviceExtension = SynchronizedParameters->DeviceExtension;

    //
    // Enable the counter to begin collecting data.
    //

    Status = EnablePerformanceCounter( deviceExtension );

    SynchronizedParameters->Status = Status;

    return;
}


VOID
StopCounterSynchronized(
    PDEVICE_EXTENSION DeviceExtension
    );

NTSTATUS
StopCounter(
    PDEVICE_EXTENSION DeviceExtension
    )
/*++

Routine Description:

    Stop the counter for the performance counter described by the
    specified device extension.

Arguments:

    DeviceExtension - Supplies a pointer to the device extension for the
                          performance counter to be stopped.

Return Value:

    The status of the operation is returned.

--*/
{
/*
    //
    // Synchronize the stop routine.
    //

    KeSynchronizeExecution( DeviceExtension->InterruptObject,
                            (PKSYNCHRONIZE_ROUTINE)StopCounterSynchronized,
                            DeviceExtension );

*/
    //
    // Disable the counter from collecting data.
    //

    DisablePerformanceCounter( DeviceExtension );

    if( DeviceExtension->Started == TRUE ){

        //
        // Theoretically, the performance counter would be stopped now
        // and so would not continue interrupting.  There are 2 problems
        // with this theory:
        //    (1)Pass 2.1 - does not have seperate performance counter disables,
        //                  the only control over the interrupts are the
        //                  interrupt enable masks
        //    (2)Pass 3 - the performance counter enable bits are or'd
        //                  together so if either is on then both performance
        //                  counters continue to interrupt
        //
        // So, we will make sure that the interrupts stop by disconnecting
        // the interrupt.  This also means that when the performance counter
        // is enabled it must reconnect if it is not already connected.
        //
        //

        IoDisconnectInterrupt( DeviceExtension->InterruptObject );

    }

    //
    // Indicate that the performance counter has been stopped.
    //

    DeviceExtension->Started = FALSE;

    return STATUS_SUCCESS;

}

VOID
StopCounterSynchronized(
    PDEVICE_EXTENSION DeviceExtension
    )
/*++

Routine Description:

    Stop the counter described by the specified device extension.

Arguments:

    DeviceExtension - Supplies a pointer to the device extension for the
                          performance counter to be stopped.

Return Value:

    None.

--*/
{

    //
    // Disable the counter from collecting data.
    //

    DisablePerformanceCounter( DeviceExtension );

    return;
}


NTSTATUS
Axp21064ReadPalCounters(
    PAXP21064_RDPALCOUNTERS PalCounters
    )
/*++

Routine Description:

    Attempt to read the PALcode event counters.  The attempt will fail
    if the rdcounters call pal is not implemented.

Arguments:

    PalCounters - Supplies a pointer to the driver formatted counters for
                      PALcode events.

Return Value:

    The status returned by AxpReadPalCounters is returned.

--*/
{
    COUNTERS_21064 Axp21064PalCounters;
    ULONG BytesRead;

    //
    // Read the PALcode event counters.
    //

    BytesRead = AxpReadPalCounters( &Axp21064PalCounters,
                                 sizeof(COUNTERS_21064) );

    switch( BytesRead ){

    //
    // Zero bytes read indicates that the call is not implemented.
    //

    case 0:

        return STATUS_NOT_IMPLEMENTED;

    //
    // Otherwise, we expect the bytes read to exactly match the sizeof
    // the 21064 counter structure.
    //

    case sizeof(COUNTERS_21064):

        break;

    //
    // The call is implemented but the expected counter structure does not
    // match.
    //

    default:

        return STATUS_INVALID_PARAMETER;

    }

    //
    // Reformat the PALcode counters to the driver-formatted counters.
    //

    //
    // Hardware vectors.
    //

    PalCounters->MachineCheckCount =
            Axp21064PalCounters.MachineCheckCount.QuadPart;
    PalCounters->ArithmeticExceptionCount =
            Axp21064PalCounters.ArithmeticExceptionCount.QuadPart;
    PalCounters->InterruptCount =
            Axp21064PalCounters.InterruptCount.QuadPart;
    PalCounters->ItbMissCount =
            Axp21064PalCounters.ItbMissCount.QuadPart;
    PalCounters->NativeDtbMissCount =
            Axp21064PalCounters.NativeDtbMissCount.QuadPart;
    PalCounters->PalDtbMissCount =
            Axp21064PalCounters.PalDtbMissCount.QuadPart;
    PalCounters->ItbAccessViolationCount =
            Axp21064PalCounters.ItbAcvCount.QuadPart;
    PalCounters->DtbAccessViolationCount =
            Axp21064PalCounters.DtbAcvCount.QuadPart;
    PalCounters->UnalignedAccessCount =
            Axp21064PalCounters.UnalignedCount.QuadPart;
    PalCounters->IllegalInstructionCount =
            Axp21064PalCounters.OpcdecCount.QuadPart;
    PalCounters->FloatingEnableExceptionCount =
            Axp21064PalCounters.FenCount.QuadPart;

    //
    // Software events.
    //

    PalCounters->ItbTranslationNotValidCount =
            Axp21064PalCounters.ItbTnvCount.QuadPart;
    PalCounters->DtbTranslationNotValidCount =
            Axp21064PalCounters.DtbTnvCount.QuadPart;
    PalCounters->PteMissCount =
            Axp21064PalCounters.PteMissCount.QuadPart;
    PalCounters->PalKernelStackMissCount =
            Axp21064PalCounters.KspMissCount.QuadPart;
    PalCounters->PdeTranslationNotValidCount =
            Axp21064PalCounters.PdeTnvCount.QuadPart;

    //
    // Privileged Call pals.
    //

    PalCounters->HaltCount =
            Axp21064PalCounters.HaltCount.QuadPart;
    PalCounters->RestartCount =
            Axp21064PalCounters.RestartCount.QuadPart;
    PalCounters->DrainAbortCount =
            Axp21064PalCounters.DrainaCount.QuadPart;
    PalCounters->InitPalCount =
            Axp21064PalCounters.InitpalCount.QuadPart;
    PalCounters->WrentryCount =
            Axp21064PalCounters.WrentryCount.QuadPart;
    PalCounters->SwapIrqlCount =
            Axp21064PalCounters.SwpirqlCount.QuadPart;
    PalCounters->ReadIrqlCount =
            Axp21064PalCounters.RdirqlCount.QuadPart;
    PalCounters->DisableInterruptCount =
            Axp21064PalCounters.DiCount.QuadPart;
    PalCounters->EnableInterruptCount =
            Axp21064PalCounters.EiCount.QuadPart;
    PalCounters->SwapPalcodeCount =
            Axp21064PalCounters.SwppalCount.QuadPart;
    PalCounters->SetSoftwareInterruptCount =
            Axp21064PalCounters.SsirCount.QuadPart;
    PalCounters->ClearSoftwareInterruptCount =
            Axp21064PalCounters.CsirCount.QuadPart;
    PalCounters->ReturnFromExceptionCount =
            Axp21064PalCounters.RfeCount.QuadPart;
    PalCounters->ReturnFromSystemCallCount =
            Axp21064PalCounters.RetsysCount.QuadPart;
    PalCounters->SwapContextCount =
            Axp21064PalCounters.SwpctxCount.QuadPart;
    PalCounters->SwapProcessCount =
            Axp21064PalCounters.SwpprocessCount.QuadPart;
    PalCounters->ReadMcesCount =
            Axp21064PalCounters.RdmcesCount.QuadPart;
    PalCounters->WriteMcesCount =
            Axp21064PalCounters.WrmcesCount.QuadPart;
    PalCounters->TbInvalidateAllCount =
            Axp21064PalCounters.TbiaCount.QuadPart;
    PalCounters->TbInvalidateSingleCount =
            Axp21064PalCounters.TbisCount.QuadPart;
    PalCounters->DtbInvalidateSingleCount =
            Axp21064PalCounters.DtbisCount.QuadPart;
    PalCounters->TbisAsnCount =
            Axp21064PalCounters.Misc1Count.QuadPart;
    PalCounters->ReadKernelStackCount =
            Axp21064PalCounters.RdkspCount.QuadPart;
    PalCounters->SwapKernelStackCount =
            Axp21064PalCounters.SwpkspCount.QuadPart;
    PalCounters->ReadProcessorStatusCount =
            Axp21064PalCounters.RdpsrCount.QuadPart;
    PalCounters->ReadPcrCount =
            Axp21064PalCounters.RdpcrCount.QuadPart;
    PalCounters->ReadThreadCount =
            Axp21064PalCounters.RdthreadCount.QuadPart;
    PalCounters->ReadCountersCount =
            Axp21064PalCounters.RdcountersCount.QuadPart;
    PalCounters->ReadStateCount =
            Axp21064PalCounters.RdstateCount.QuadPart;
    PalCounters->WritePerfmonCount =
            Axp21064PalCounters.WrperfmonCount.QuadPart;
    PalCounters->InitPcrCount =
            Axp21064PalCounters.InitpcrCount.QuadPart;

    //
    // Unprivileged Call pals.
    //

    PalCounters->UserBreakpointCount =
            Axp21064PalCounters.BptCount.QuadPart;
    PalCounters->SystemCallCount =
            Axp21064PalCounters.CallsysCount.QuadPart;
    PalCounters->ImbCount =
            Axp21064PalCounters.ImbCount.QuadPart;
    PalCounters->GenerateTrapCount =
            Axp21064PalCounters.GentrapCount.QuadPart;
    PalCounters->ReadTebCount =
            Axp21064PalCounters.RdtebCount.QuadPart;
    PalCounters->KernelBreakpointCount =
            Axp21064PalCounters.KbptCount.QuadPart;
    PalCounters->CallKernelDebuggerCount =
            Axp21064PalCounters.CallkdCount.QuadPart;

    //
    // Miscellaneous counters.
    //

    PalCounters->Misc1Count = Axp21064PalCounters.Misc2Count.QuadPart;
    PalCounters->Misc2Count = Axp21064PalCounters.Misc3Count.QuadPart;
    PalCounters->Misc3Count = Axp21064PalCounters.Misc4Count.QuadPart;

    return STATUS_SUCCESS;

}


NTSTATUS
EnablePerformanceCounter(
    PDEVICE_EXTENSION DeviceExtension
    )
/*++

Routine Description:

    Enable the performance counter to begin counting events using the
    current event count and mux control from the specificed device
    extension.

    N.B. - Expected to be synchronized with respect to the interrupt
           handler and any other writers of the device extension.
Arguments:

    DeviceExtension - Supplies a pointer to the device extension for the
                          performance counter to be enabled.

Return Value:

    The status of the operation is returned.

--*/
{
    EVENT_COUNT EventCount;
//    NTSTATUS Status;

    //
    // Translate the event count to its symbolic value.
    //

    if( DeviceExtension->Counter == PerformanceCounter0 ){
        if( DeviceExtension->EventCount == CountEvents2xx12 ){
            EventCount = EventCountHigh;
        } else {
            EventCount = EventCountLow;
        }
    }

    if( DeviceExtension->Counter == PerformanceCounter1 ){
        if( DeviceExtension->EventCount == CountEvents2xx8 ){
            EventCount = EventCountHigh;
        } else {
            EventCount = EventCountLow;
        }
    }

    //
    // Write the performance counter with the new values.
    //

    AxpWritePerformanceCounter( (ULONG)DeviceExtension->Counter,
                                (ULONG)CounterEnable,
                                (ULONG)DeviceExtension->MuxControl,
                                (ULONG)EventCount );

    return STATUS_SUCCESS;
}

VOID
DisablePerformanceCounter(
    PDEVICE_EXTENSION DeviceExtension
    )
/*++

Routine Description:

    Disable the performance counter from counting events.

    N.B. - Expected to be synchronized with respect to the interrupt
           handler and any other writers of the device extension.

Arguments:

    DeviceExtension - Supplies a pointer to the device extension for the
                          performance counter to be disabled.

Return Value:

    None.

--*/
{

    //
    // Disable the performance counter.
    //

    if( DeviceExtension->Started == TRUE ){

        //
        // Disable the performance counter.
        //

        AxpWritePerformanceCounter( (ULONG)DeviceExtension->Counter,
                                    (ULONG)CounterDisable,
                                    (ULONG)0,
                                    (ULONG)0 );
    }

    return;
}

