//      TITLE("Support routines for AXP performance counter driver")
//++
//
// Copyright (c) 1993  Digital Equipment Corporation
//
// Module Name:
//
//    axperf.s
//
// Abstract:
//
//    This module implements the low-level routines that interface directly
//    to the privileged call pal instructions that access the performance
//    and event counters.
//
// Author:
//
//    Joe Notarangelo  27-Jun-1993
//
// Environment:
//
//    Kernel mode only.
//
// Revision History:
//
//--

#include "ksalpha.h"

//++
//
// ULONG
// AxpReadPalCounters(
//    IN PVOID CounterBuffer,
//    IN ULONG BufferLength
//    )
//
// Routine Description:
//
//    This routine interfaces with the call pal instruction: rdcounters.
//
// Arguments:
//
//    CounterBuffer (a0) - Supplies a pointer to the buffer to receive the
//                             the PALcode counter data.
//    BufferLength (a1) - Supplies the length of the buffer.
//
// Return Value:
//
//    The number of bytes read by the call pal is returned.
//
//--

	LEAF_ENTRY(AxpReadPalCounters)

	call_pal rdcounters             // read the counters, v0 = result

	ret     zero, (ra)              // return

	.end AxpReadPalCounters


//++
// VOID
// WritePerformanceCounter(
//     IN ULONG PerformanceCounter,
//     IN ULONG Enable,
//     IN ULONG MuxControl OPTIONAL,
//     IN ULONG EventCount OPTIONAL
//     )
//
// Routine Description:
//
//    This routine interfaces with the call pal instruction: wrperfmon.
//
// Arguments:
//
//    PerformanceCounter - Supplies the number of the performance counter to be
//                             written.
//    Enable - Supplies a 1 if the counter is to be enabled and a 0 if the
//                 counter is to be disabled.
//    MuxControl - Supplies the mux control value for the counter (iff enable).
//    EventCount - Supplies the event count value for the counter (iff enable).
//
// Return Value:
//
//    None.
//
//--

	LEAF_ENTRY( AxpWritePerformanceCounter )

	call_pal wrperfmon              // write the counter

	ret     zero, (ra)              // return

	.end AxpWritePerformanceCounter



//++
// PVOID
// GetProcessorPCR(
//     )
//
// Routine Description:
//
//    This routine interfaces with the call pal instruction: rdpcr
//
// Arguments:
//
//    None

// Return Value:
//
//    PCR base address.
//
//--

	LEAF_ENTRY( GetProcessorPCR )

	call_pal rdpcr                  // read processor PCR, v0 is the PCR base addr

	ret     zero, (ra)              // return it

	.end GetProcessorPCR


//++
// PVOID
// GetCurTEB(
//     )
//
// Routine Description:
//
//    This routine interfaces with the call pal instruction: rdteb
//
// Arguments:
//
//    None

// Return Value:
//
//    Thread base address.
//
//--

	LEAF_ENTRY( GetCurTEB )

	call_pal rdteb                  // return addr. to current TEB in v0

	ret     zero, (ra)              // return it

	.end GetCurTEB


//++
// PVOID
// GetCurThread(
//     )
//
// Routine Description:
//
//    This routine interfaces with the call pal instruction: rdthread
//
// Arguments:
//
//    None

// Return Value:
//
//    Thread object addr (PCR?).
//
//--

	LEAF_ENTRY( GetCurThread )

	call_pal rdthread                       // return addr. to current TEB in v0

	ret     zero, (ra)              // return it

	.end GetCurThread


//++
// PVOID
// GetCurrentStackLoc(VOID
//     )
//
// Routine Description:
//
//    This routine simply returns the stack pointer on entry to
//    this routine (the caller is responsible for subtracting
//    the size of the call frame).
//
// Arguments:
//
//    None

// Return Value:
//
//    VA pointed to by stack (callers must subtract size of call fram).
//
//--

	LEAF_ENTRY( GetCurrentStackLoc )

	mov     sp, v0
	ret     zero , (ra)             // return

	.end GetCurrentStackLoc



//++
// PVOID
// GetCurPID(VOID
//     )
//
// Routine Description:
//
//    This routine simply returns the current process id
//
// Arguments:
//
//    None
//
// Return Value:
//
//    PID pointed to by global KiCurrentProcess
//
//--
//
//      LEAF_ENTRY( GetCurPID )
//
//      lda     v0, KiCurrentProcess
//      ret     zero , (ra)             // return
//
//      .end GetCurPID

//++
// ULONG
// ExInterlockedVerifyList(
//     IN PLIST_ENTRY ListHead
//     )
//
// Routine Description:
//
//    This routine verifies a double-linked list
//
// Arguments:
//
//    ListHead (a0) - Supplies a pointer to the head of the doubly link
//	list which is to be verifies
//
// Return Value:
//
//    (v0) Number of elements on list (not including header) if list ok
//		*** If list not ok, routine will crash ***
//
//--
//

	LEAF_ENTRY( ExInterlockedVerifyList )

	DISABLE_INTERRUPTS

	// First, follow flinks around
		
	mov	a0, t0			// t0 is roving ptr
	subq	zero, 1, t2		// start count at -1 to skip list head
	subq	zero, 1, t3		// ditto for blink count

flnext:	
	ldl	t0, LsFlink(t0)		// follow link
	cmpeq	t0, a0, t1		// compare to head (t1==1 -> at head) 
	addq	t2, 1, t2		// t2 has number of Flink elements
	beq	t1, flnext		// br if not at head

	// ok, next follow blinks the other way
	
	// t0 == a0 here because of beq
	
blprev:
	ldl	t0, LsBlink(t0)
	cmpeq	t0, a0, t1		// compare to head (t1==1 -> at head)
	addq	t3, 1, t3		// t3 has number of Blink elements
	beq	t1, blprev		// br if not at head

	// t2 (from flinks) should equal t3 (from blinks)
	
	cmpeq	t2, t3, t4
	beq	t4, bug
	
	ENABLE_INTERRUPTS

	mov	t2, v0
	ret	zero, (ra)

bug:
	call_pal	halt
	br	zero, bug
		
	.end ExInterLockedVerifyList

	LEAF_ENTRY( PCInterruptAddFrame )

	mov	fp, a2

	ldl	v0, __imp_PerformanceCounterInterrupt
	jmp	zero, (v0)

	.end	PCInterruptAddFrame
