/*++

Copyright (c) 1993  Digital Equipment Corporation

Module Name:

    perfsup.c

Abstract:

    This module implements support functions for the Alpha AXP 21064
    performance counter driver.

Author:

    Joe Notarangelo  26-Jun-1993

Environment:

    Kernel mode only.

Notes:


Revision History:

    Brad Waters       1-Sep-1994
    Added include of ints.h to defined uint64 which was added to the
    DEVICE_EXTENSION structure in perfev4.h.  Also added calls to trace
    facility.
--*/
//#include "ints.h"
#include "ntddk.h"
#include "dda21064.h"
#include "ipr_kernel_int_defs.h"

#include "ipr_driver_nt.h"

#include "trpfrm.h"
#include "perfev4.h"
#include "ev5_masks_common.h"
//#include "TraceDrv.h"

extern void PCInterruptAddFrame(void);

typedef enum {StoreDeviceObject, RetrieveDeviceObject} DeviceObjectOperation;
typedef enum
{
   //
   // Counter 0 events
   //
   EV5_MUX0_CYCLES,
   EV5_MUX0_ISSUES,
   EV5_MUX0_TOTAL
} EV5_MUX0;

//
// Counter 1 events
//
typedef enum
{
   EV5_MUX1_NON_ISSUE_CYCLES,
   EV5_MUX1_SPLIT_ISSUE_CYCLES,
   EV5_MUX1_PIPELINE_DRY,
   EV5_MUX1_REPLAY_TRAP,
   EV5_MUX1_SINGLE_ISSUE_CYCLES,
   EV5_MUX1_DUAL_ISSUE_CYCLES,
   EV5_MUX1_TRIPLE_ISSUE_CYCLES,
   EV5_MUX1_QUAD_ISSUE_CYCLES,
   EV5_MUX1_FLOW_CHANGE,
   EV5_MUX1_INTEGER_OPERATE,
   EV5_MUX1_FP_INSTRUCTIONS,
   EV5_MUX1_LOAD_INSTRUCTIONS,
   EV5_MUX1_STORE_INSTRUCTIONS,
   EV5_MUX1_ICACHE_ACCESS,
   EV5_MUX1_DCACHE_ACCESS,
   EV5_MUX1_CBOX1,
   EV5_MUX1_TOTAL
} EV5_MUX1;

//
// Counter 1 CBOX events
//
typedef enum
{
   EV5_CBOX1_SCACHE_ACCESS,
   EV5_CBOX1_SCACHE_READ,
   EV5_CBOX1_SCACHE_WRITE,
   EV5_CBOX1_SCACHE_VICTIM,
   EV5_CBOX1_RESERVED4,
   EV5_CBOX1_BCACHE_HIT,
   EV5_CBOX1_BCACHE_VICTIM,
   EV5_CBOX1_SYS_REQ,
   EV5_CBOX1_TOTAL
} EV5_CBOX1;

typedef enum
{
   PCA56_CBOX1_BCACHE_READ,
   PCA56_CBOX1_BCACHE_D_READ_HITS,
   PCA56_CBOX1_SCACHE_D_READ_FILLS,
   PCA56_CBOX1_BCACHE_WRITES,
   PCA56_CBOX1_RESERVED4,
   PCA56_CBOX1_BCACHE_CLEAN_WRITE_HITS,
   PCA56_CBOX1_BCACHE_VICTIM,
   PCA56_CBOX1_READ_MISS_2_LAUNCHED,
   PCA56_CBOX1_TOTAL
} PCA56_CBOX1;

//
// Counter 2 events
//
typedef enum
{
   EV5_MUX2_LONG_STALLS,
   EV5_MUX2_RESERVED1,
   EV5_MUX2_PC_MISPR,
   EV5_MUX2_BRANCH_MISPREDICTS,
   EV5_MUX2_ICACHE_MISSES,
   EV5_MUX2_ITB_MISS,
   EV5_MUX2_DCACHE_MISSES,
   EV5_MUX2_DTB_MISS,
   EV5_MUX2_LOADS_MERGED,
   EV5_MUX2_LDU_REPLAYS,
   EV5_MUX2_WB_MAF_FULL_REPLAYS,
   EV5_MUX2_EXTERNAL,
   EV5_MUX2_CYCLES,
   EV5_MUX2_MEM_BARRIER,
   EV5_MUX2_LOAD_LOCKED,
   EV5_MUX2_CBOX2,
   EV5_MUX2_TOTAL
} EV5_MUX2;

//
// Counter 2 CBOX events
//
typedef enum
{
   EV5_CBOX2_SCACHE_MISS,
   EV5_CBOX2_SCACHE_READ_MISS,
   EV5_CBOX2_SCACHE_WRITE_MISS,
   EV5_CBOX2_SCACHE_SH_WRITE,
   EV5_CBOX2_SCACHE_WRITE,
   EV5_CBOX2_BCACHE_MISS,
   EV5_CBOX2_SYS_INV,
   EV5_CBOX2_SYS_READ_REQ,
   EV5_CBOX2_TOTAL
} EV5_CBOX2;

typedef enum
{
   PCA56_CBOX2_BCACHE_D_READS,
   PCA56_CBOX2_BCACHE_READ_HITS,
   PCA56_CBOX2_BCACHE_READ_FILLS,
   PCA56_CBOX2_BCACHE_WRITE_HITS,
   PCA56_CBOX2_BCACHE_WRITE_FILLS,
   PCA56_CBOX2_SYS_READ_FLUSH_BCACHE_HITS,
   PCA56_CBOX2_SYS_READ_FLUSH_BCACHE_MISSES,
   PCA56_CBOX2_READ_MISS_3_LAUNCHED,
   PCA56_CBOX2_TOTAL
} PCA56_CBOX2;

#define EV5_FREQ_M_LOW 0
#define EV5_FREQ_M_HIGH 1
#define EV5_FREQ_M_INTERRUPT 2

NTSTATUS
EnablePerformanceCounter(
    PDEVICE_EXTENSION DeviceExtension
    );

VOID
DisablePerformanceCounter(
    PDEVICE_EXTENSION DeviceExtension
    );
unsigned long int DeriveShiftValue(unsigned long int mask);


BOOLEAN
ValidateEventCount(
    PAXP21064_SETCOUNTER EventCountParameters,
    AXP21064_PCCOUNTER PerformanceCounter
    )
/*++

Routine Description:

    This routine validates the parameters for a set event count ioctl
    are valid.

Arguments:

    EventCountParameters - Supplies a pointer to the event count parameters
                               to be validated.

    PerformanceCounter - Supplies the performance counter for which the
                             parameters are to be validated.

Return Value:

    TRUE is returned if the parameters are valid, otherwise false is
    returned.

--*/
{

    //
    // Validate the event count values as appropriate for the
    // given performance counter.
    //

    switch( PerformanceCounter ){

    //
    // Handle Performance Counter 0
    //

    case PerformanceCounter0:

        switch( EventCountParameters->EventCount ){

        //
        // The valid cases.
        //

        case CountEvents2xx12:
        case CountEvents2xx16:

            return TRUE;

        //
        // Otherwise, not valid.
        //

        default:

            return FALSE;

        } //end switch EventCountParameters->EventCount

    //
    // Handle Performance Counter 1
    //

    case PerformanceCounter1:

        switch( EventCountParameters->EventCount ){

        //
        // The valid cases.
        //

        case CountEvents2xx8:
        case CountEvents2xx12:

            return TRUE;

        default:

            return FALSE;
        } //end switch EventCountParameters->EventCount

    //
    // Handle Performance Counter 2
    //

    case PerformanceCounter2:

        switch( EventCountParameters->EventCount ){

        //
        // The valid cases.
        //

        case CountEvents2xx8:
        case CountEvents2xx12:

            return TRUE;


        //
        // Otherwise, not valid.
        //

        default:

            return FALSE;

        } //end switch EventCountParameters->EventCount

    //
    // Otherwise, invalid performance counter
    //

    default:

        return FALSE;

    } //end switch PerformanceCounter

}


//
// Define structure for passing parameters to the synchronized set
// routine.
//

typedef struct _EVENTCOUNT_SYNCH{
    PDEVICE_EXTENSION DeviceExtension;
    PAXP21064_SETCOUNTER EventCountParameters;
    NTSTATUS Status;
} EVENTCOUNT_SYNCH, *PEVENTCOUNT_SYNCH;

VOID
SetEventCountSynchronized(
    PEVENTCOUNT_SYNCH SynchronizedParameters
    );

NTSTATUS SetEventCount(PDEVICE_EXTENSION DeviceExtension, uint64 *EventCountParameters)
/*++

Routine Description:

    This routine sets a new event count for the performance counter
    described by the specified device extension.

Arguments:

    DeviceExtension - Supplies a pointer to the device extension for the
                          performance counter to be modified.

    EventCountParameters - Supplies a pointer to the new event count
                               parameters for the performance counter.

Return Value:

    The status from the synchronized update routine is returned.

--*/
{
    DeviceExtension->EventCount = *EventCountParameters;
    return STATUS_SUCCESS;
}

VOID
SetEventCountSynchronized(
    PEVENTCOUNT_SYNCH SynchronizedParameters
    )
/*++

Routine Description:

    This routine does the actual work of setting new event count parameters
    for a performance counter.  It is synchronized with the interrupt
    routine and all other writers of the device extension.

Arguments:

    SynchronizedParameters - Supplies the data used to set a new event count
                                 for a performance counter.

Return Value:

    None.  However, status is returned via the SynchronizedParameters
    structure.

--*/
{
    PDEVICE_EXTENSION deviceExtension;
    PAXP21064_SETCOUNTER EventCountParameters;
//    EVENT_COUNT EventCount;
    NTSTATUS Status;

    deviceExtension = SynchronizedParameters->DeviceExtension;
    EventCountParameters = SynchronizedParameters->EventCountParameters;

    //
    // Set the new event count in the device extension.
    //

    deviceExtension->EventCount = EventCountParameters->EventCount;

    //
    // If the counter is current started, then enable the counter to
    // begin collecting data.
    //

    Status = EnablePerformanceCounter( deviceExtension );

    SynchronizedParameters->Status = Status;

    return;

}


BOOLEAN
ValidateMuxControl(
    PAXP21064_SETMUXCONTROL MuxControlParameters,
    AXP21064_PCCOUNTER PerformanceCounter
    )
/*++

Routine Description:

    This routine validates the parameters for a set mux control ioctl
    are valid.

Arguments:

    MuxControlParameters - Supplies a pointer to the mux control parameters
                               to be validated.

    PerformanceCounter - Supplies the performance counter for which the
                             the mux control parameters must be validated.
Return Value:

    TRUE is returned if the parameters are valid, otherwise false is
    returned.

--*/
{

    //
    // Validate the event count values as appropriate for the
    // given performance counter.
    //

    switch( PerformanceCounter ){

    //
    // Handle Performance Counter 0
    //

    case PerformanceCounter0:

        switch( MuxControlParameters->MuxControl ){

        //
        // The valid cases.
        //

        case TotalIssues:
        case PipelineDry:
        case LoadInstruction:
        case PipelineFrozen:
        case BranchInstructions:
        case PalMode:
        case TotalCycles:
        case TotalNonIssues:
        case ExternalCounter0:

            return TRUE;

        //
        // Otherwise, not valid.
        //

        default:

            return FALSE;

        } //end switch MuxControlParameters->MuxControl

    //
    // Handle Performance Counter 1
    //

    case PerformanceCounter1:

        switch( MuxControlParameters->MuxControl ){

        //
        // The valid cases.
        //

        case DcacheMiss:
        case IcacheMiss:
        case DualIssues:
        case BranchMispredicts:
        case FPInstructions:
        case IntegerOperate:
        case StoreInstructions:
        case ExternalCounter1:

            return TRUE;

        //
        // Otherwise, not valid.
        //

        default:

            return FALSE;
        } //end switch MuxControlParameters->MuxControl
//++++
    //
    // Handle Performance Counter 2
    //

    case PerformanceCounter2:

        switch( MuxControlParameters->MuxControl ){

        //
        // The valid cases.
        //

        case DcacheMiss:
        case IcacheMiss:
        case DualIssues:
        case BranchMispredicts:
        case FPInstructions:
        case IntegerOperate:
        case StoreInstructions:
        case ExternalCounter1:

            return TRUE;

        //
        // Otherwise, not valid.
        //

        default:

            return FALSE;


        } //end switch MuxControlParameters->MuxControl

    //
    // Otherwise, invalid performance counter
    //

    default:

        return FALSE;

    } //end switch PerformanceCounter

}


//
// Define structure for passing parameters to the synchronized set
// routine.
//

typedef struct _MUXCONTROL_SYNCH{
    PDEVICE_EXTENSION DeviceExtension;
    PAXP21064_SETMUXCONTROL MuxControlParameters;
    NTSTATUS Status;
} MUXCONTROL_SYNCH, *PMUXCONTROL_SYNCH;

VOID
SetMuxControlSynchronized(
    PMUXCONTROL_SYNCH SynchronizedParameters
    );

NTSTATUS SetMuxControl(PDEVICE_EXTENSION DeviceExtension, uint64 *MuxControlParameters)
/*++

Routine Description:

    This routine sets a new mux control for the performance counter
    described by the specified device extension.

Arguments:

    DeviceExtension - Supplies a pointer to the device extension for the
                          performance counter to be modified.

    MuxControlParameters - Supplies a pointer to the new mux control
                               parameters for the performance counter.

Return Value:

    The status of the synchronized update routine is returned.

--*/
{
    //
    // Set the new mux control in the device extension.
    //

    DeviceExtension->MuxControl = *MuxControlParameters;

    return STATUS_SUCCESS;

}

VOID
SetMuxControlSynchronized(
    PMUXCONTROL_SYNCH SynchronizedParameters
    )
/*++

Routine Description:

    This routine does the actual work of setting new mux control parameters
    for a performance counter.  It is synchronized with the interrupt
    routine and all other writers of the device extension.

Arguments:

    SynchronizedParameters - Supplies the data used to set a new mux control
                                 for a performance counter.

Return Value:

    None.  However, the status of the operation is returned via the
    SynchronizedParameters structure.

--*/
{
    PDEVICE_EXTENSION deviceExtension;
    PAXP21064_SETMUXCONTROL MuxControlParameters;
//    EVENT_COUNT EventCount;
    NTSTATUS Status;


    deviceExtension = SynchronizedParameters->DeviceExtension;
    MuxControlParameters = SynchronizedParameters->MuxControlParameters;

    //
    // Set the new mux control in the device extension.
    //

    deviceExtension->MuxControl = MuxControlParameters->MuxControl;

    //
    // If the counter is current started, then enable the counter to
    // begin collecting data.
    //


    Status = EnablePerformanceCounter( deviceExtension );


    SynchronizedParameters->Status = Status;

    return;

}


//
// Define structure for passing parameters to the synchronized set
// routine.
//

typedef struct _STARTCOUNTER_SYNCH{
    PDEVICE_EXTENSION DeviceExtension;
    NTSTATUS Status;
} STARTCOUNTER_SYNCH, *PSTARTCOUNTER_SYNCH;

VOID
StartCounterSynchronized(
    PSTARTCOUNTER_SYNCH SynchronizedParameters
    );

NTSTATUS StartCounter(PDEVICE_EXTENSION DeviceExtension)
//
//
//Routine Description:
//
//    Start the counter for the performance counter described by the
//    specified device extension.
//
//Arguments:
//
//    DeviceExtension - Supplies a pointer to the device extension for the
//                          performance counter to be started.
//
//Return Value:
//
//    The status of the operation is returned.
//
//
{
	NTSTATUS Status;
	ULONG InterruptingCounters = DeviceExtension->EventCount;
	unsigned long int CounterNumber = 2;
	DEVICE_OBJECT *pDO;
	DEVICE_EXTENSION *pDE;

	InterruptingCounters >>= 4;

	while(InterruptingCounters != 0)
	  {
		if ((InterruptingCounters & 0x3) & EV5_FREQ_M_INTERRUPT)
		  {
//TrcPrint("Starting counter %d on cpu %d\n", CounterNumber, DeviceExtension->cpu);
			pDO = GetSetDeviceObject(RetrieveDeviceObject,
			  DeviceExtension->cpu,
			  CounterNumber,
			  NULL);
			pDE = pDO->DeviceExtension;

			Status = IoConnectInterrupt(
			  &pDE->InterruptObject,
			  (PKSERVICE_ROUTINE)PCInterruptAddFrame,
			  (PVOID)pDE,
			  pDE->ISRSpin,
			  pDE->Vector,
			  pDE->Irql,
			  pDE->Irql,
			  Latched,
			  TRUE,
			  pDE->Affinity,
			  FALSE );

			if( Status == STATUS_SUCCESS )
			  {
				pDE->Started = TRUE;
//TrcPrint("Counter started\n");
			  }
			else
			  {
				AxpPrint((1, "Error reconnecting interrupt Status = %x\n", Status ));
				return Status;
			  }
		  }

		InterruptingCounters >>= 2;
		CounterNumber--;
	  }

	Status = EnablePerformanceCounter( DeviceExtension );

	if( !NT_SUCCESS(Status) )
	  {
		return Status;
	  }


    return Status;
}

VOID
StartCounterSynchronized(
    PSTARTCOUNTER_SYNCH SynchronizedParameters
    )
/*++

Routine Description:

    Start the counter described by the specified device extension.

Arguments:

    SynchronizedParameters - Supplies a pointer to the synchronized
                                 parameters for the start operation.

Return Value:

    None.  However, the a status is returned via the SynchronizedParameters.

--*/
{
    PDEVICE_EXTENSION deviceExtension;
    NTSTATUS Status;

    deviceExtension = SynchronizedParameters->DeviceExtension;

    //
    // Enable the counter to begin collecting data.
    //

    Status = EnablePerformanceCounter( deviceExtension );

    SynchronizedParameters->Status = Status;

    return;
}


VOID
StopCounterSynchronized(
    PDEVICE_EXTENSION DeviceExtension
    );

NTSTATUS StopCounter(PDEVICE_EXTENSION DeviceExtension)
//
//
//Routine Description:
//
//    Stop the counter for the performance counter described by the
//    specified device extension.
//
//Arguments:
//
//    DeviceExtension - Supplies a pointer to the device extension for the
//                          performance counter to be stopped.
//
//Return Value:
//
//    The status of the operation is returned.
//
//
  {
	ULONG InterruptingCounters = DeviceExtension->EventCount;
	unsigned long int CounterNumber = 2;
	DEVICE_OBJECT *pDO;
	DEVICE_EXTENSION *pDE;

	DisablePerformanceCounter(DeviceExtension);

	InterruptingCounters >>= 4;

	while(InterruptingCounters != 0)
	  {
		if ((InterruptingCounters & 0x3) & EV5_FREQ_M_INTERRUPT)
		  {
//TrcPrint("Stopping counter %d on cpu %d\n", CounterNumber, DeviceExtension->cpu);
			pDO = GetSetDeviceObject(RetrieveDeviceObject,
			  DeviceExtension->cpu,
			  CounterNumber,
			  NULL);
			pDE = pDO->DeviceExtension;

		IoDisconnectInterrupt(pDE->InterruptObject);

			pDE->Started = FALSE;
		  }

		InterruptingCounters >>= 2;
		CounterNumber--;
	  }

    return STATUS_SUCCESS;
  }

VOID
StopCounterSynchronized(
    PDEVICE_EXTENSION DeviceExtension
    )
/*++

Routine Description:

    Stop the counter described by the specified device extension.

Arguments:

    DeviceExtension - Supplies a pointer to the device extension for the
                          performance counter to be stopped.

Return Value:

    None.

--*/
{

    //
    // Disable the counter from collecting data.
    //

    DisablePerformanceCounter( DeviceExtension );

    return;
}


NTSTATUS
Axp21064ReadPalCounters(
    PAXP21064_RDPALCOUNTERS PalCounters
    )
/*++

Routine Description:

    Attempt to read the PALcode event counters.  The attempt will fail
    if the rdcounters call pal is not implemented.

Arguments:

    PalCounters - Supplies a pointer to the driver formatted counters for
                      PALcode events.

Return Value:

    The status returned by AxpReadPalCounters is returned.

--*/
{
    COUNTERS_21064 Axp21064PalCounters;
    ULONG BytesRead;

    //
    // Read the PALcode event counters.
    //

    BytesRead = AxpReadPalCounters( &Axp21064PalCounters,
                                 sizeof(COUNTERS_21064) );

    switch( BytesRead ){

    //
    // Zero bytes read indicates that the call is not implemented.
    //

    case 0:

        return STATUS_NOT_IMPLEMENTED;

    //
    // Otherwise, we expect the bytes read to exactly match the sizeof
    // the 21064 counter structure.
    //

    case sizeof(COUNTERS_21064):

        break;

    //
    // The call is implemented but the expected counter structure does not
    // match.
    //

    default:

        return STATUS_INVALID_PARAMETER;

    }

    //
    // Reformat the PALcode counters to the driver-formatted counters.
    //

    //
    // Hardware vectors.
    //

    PalCounters->MachineCheckCount =
            Axp21064PalCounters.MachineCheckCount.QuadPart;
    PalCounters->ArithmeticExceptionCount =
            Axp21064PalCounters.ArithmeticExceptionCount.QuadPart;
    PalCounters->InterruptCount =
            Axp21064PalCounters.InterruptCount.QuadPart;
    PalCounters->ItbMissCount =
            Axp21064PalCounters.ItbMissCount.QuadPart;
    PalCounters->NativeDtbMissCount =
            Axp21064PalCounters.NativeDtbMissCount.QuadPart;
    PalCounters->PalDtbMissCount =
            Axp21064PalCounters.PalDtbMissCount.QuadPart;
    PalCounters->ItbAccessViolationCount =
            Axp21064PalCounters.ItbAcvCount.QuadPart;
    PalCounters->DtbAccessViolationCount =
            Axp21064PalCounters.DtbAcvCount.QuadPart;
    PalCounters->UnalignedAccessCount =
            Axp21064PalCounters.UnalignedCount.QuadPart;
    PalCounters->IllegalInstructionCount =
            Axp21064PalCounters.OpcdecCount.QuadPart;
    PalCounters->FloatingEnableExceptionCount =
            Axp21064PalCounters.FenCount.QuadPart;

    //
    // Software events.
    //

    PalCounters->ItbTranslationNotValidCount =
            Axp21064PalCounters.ItbTnvCount.QuadPart;
    PalCounters->DtbTranslationNotValidCount =
            Axp21064PalCounters.DtbTnvCount.QuadPart;
    PalCounters->PteMissCount =
            Axp21064PalCounters.PteMissCount.QuadPart;
    PalCounters->PalKernelStackMissCount =
            Axp21064PalCounters.KspMissCount.QuadPart;
    PalCounters->PdeTranslationNotValidCount =
            Axp21064PalCounters.PdeTnvCount.QuadPart;

    //
    // Privileged Call pals.
    //

    PalCounters->HaltCount =
            Axp21064PalCounters.HaltCount.QuadPart;
    PalCounters->RestartCount =
            Axp21064PalCounters.RestartCount.QuadPart;
    PalCounters->DrainAbortCount =
            Axp21064PalCounters.DrainaCount.QuadPart;
    PalCounters->InitPalCount =
            Axp21064PalCounters.InitpalCount.QuadPart;
    PalCounters->WrentryCount =
            Axp21064PalCounters.WrentryCount.QuadPart;
    PalCounters->SwapIrqlCount =
            Axp21064PalCounters.SwpirqlCount.QuadPart;
    PalCounters->ReadIrqlCount =
            Axp21064PalCounters.RdirqlCount.QuadPart;
    PalCounters->DisableInterruptCount =
            Axp21064PalCounters.DiCount.QuadPart;
    PalCounters->EnableInterruptCount =
            Axp21064PalCounters.EiCount.QuadPart;
    PalCounters->SwapPalcodeCount =
            Axp21064PalCounters.SwppalCount.QuadPart;
    PalCounters->SetSoftwareInterruptCount =
            Axp21064PalCounters.SsirCount.QuadPart;
    PalCounters->ClearSoftwareInterruptCount =
            Axp21064PalCounters.CsirCount.QuadPart;
    PalCounters->ReturnFromExceptionCount =
            Axp21064PalCounters.RfeCount.QuadPart;
    PalCounters->ReturnFromSystemCallCount =
            Axp21064PalCounters.RetsysCount.QuadPart;
    PalCounters->SwapContextCount =
            Axp21064PalCounters.SwpctxCount.QuadPart;
    PalCounters->SwapProcessCount =
            Axp21064PalCounters.SwpprocessCount.QuadPart;
    PalCounters->ReadMcesCount =
            Axp21064PalCounters.RdmcesCount.QuadPart;
    PalCounters->WriteMcesCount =
            Axp21064PalCounters.WrmcesCount.QuadPart;
    PalCounters->TbInvalidateAllCount =
            Axp21064PalCounters.TbiaCount.QuadPart;
    PalCounters->TbInvalidateSingleCount =
            Axp21064PalCounters.TbisCount.QuadPart;
    PalCounters->DtbInvalidateSingleCount =
            Axp21064PalCounters.DtbisCount.QuadPart;
    PalCounters->TbisAsnCount =
            Axp21064PalCounters.Misc1Count.QuadPart;
    PalCounters->ReadKernelStackCount =
            Axp21064PalCounters.RdkspCount.QuadPart;
    PalCounters->SwapKernelStackCount =
            Axp21064PalCounters.SwpkspCount.QuadPart;
    PalCounters->ReadProcessorStatusCount =
            Axp21064PalCounters.RdpsrCount.QuadPart;
    PalCounters->ReadPcrCount =
            Axp21064PalCounters.RdpcrCount.QuadPart;
    PalCounters->ReadThreadCount =
            Axp21064PalCounters.RdthreadCount.QuadPart;
    PalCounters->ReadCountersCount =
            Axp21064PalCounters.RdcountersCount.QuadPart;
    PalCounters->ReadStateCount =
            Axp21064PalCounters.RdstateCount.QuadPart;
    PalCounters->WritePerfmonCount =
            Axp21064PalCounters.WrperfmonCount.QuadPart;
    PalCounters->InitPcrCount =
            Axp21064PalCounters.InitpcrCount.QuadPart;

    //
    // Unprivileged Call pals.
    //

    PalCounters->UserBreakpointCount =
            Axp21064PalCounters.BptCount.QuadPart;
    PalCounters->SystemCallCount =
            Axp21064PalCounters.CallsysCount.QuadPart;
    PalCounters->ImbCount =
            Axp21064PalCounters.ImbCount.QuadPart;
    PalCounters->GenerateTrapCount =
            Axp21064PalCounters.GentrapCount.QuadPart;
    PalCounters->ReadTebCount =
            Axp21064PalCounters.RdtebCount.QuadPart;
    PalCounters->KernelBreakpointCount =
            Axp21064PalCounters.KbptCount.QuadPart;
    PalCounters->CallKernelDebuggerCount =
            Axp21064PalCounters.CallkdCount.QuadPart;

    //
    // Miscellaneous counters.
    //

    PalCounters->Misc1Count = Axp21064PalCounters.Misc2Count.QuadPart;
    PalCounters->Misc2Count = Axp21064PalCounters.Misc3Count.QuadPart;
    PalCounters->Misc3Count = Axp21064PalCounters.Misc4Count.QuadPart;

    return STATUS_SUCCESS;

}

NTSTATUS EnablePerformanceCounter(PDEVICE_EXTENSION DeviceExtension)
//
//
//Routine Description:
//
//    Enable the performance counter to begin counting events using the
//    current event count and mux control from the specificed device
//    extension.
//
//    N.B. - Expected to be synchronized with respect to the interrupt
//           handler and any other writers of the device extension.
//Arguments:
//
//    DeviceExtension - Supplies a pointer to the device extension for the
//                          performance counter to be enabled.
//
//Return Value:
//
//    The status of the operation is returned.
//
//
  {
	ULONG cbox1, cbox2, RawMux, RawFctl, RawModes, CounterValue;

	RawMux = DeviceExtension->MuxControl;
	RawFctl = DeviceExtension->EventCount;
	RawModes = DeviceExtension->Modes;

	CounterValue = RawMux & ~(EV5_MUXCTL_M_CBOX1 | EV5_MUXCTL_M_CBOX2);
	CounterValue |= ((RawFctl & (EV5_FCTL_M_PCFREQ2 | EV5_FCTL_M_PCFREQ1 | EV5_FCTL_M_PCFREQ0)) << 6);
	CounterValue |= (RawModes & (EV5_MODESEL_M_KILL_KERNEL | EV5_MODESEL_M_KILL_PAL | EV5_MODESEL_M_KILL_USER));

	cbox1 = (RawMux & EV5_MUXCTL_M_CBOX1) >> DeriveShiftValue(EV5_MUXCTL_M_CBOX1);
	cbox2 = (RawMux & EV5_MUXCTL_M_CBOX2) >> DeriveShiftValue(EV5_MUXCTL_M_CBOX2);

//TrcPrint("RAW: freq: %x mux: %x modes: %x\n", RawFctl, RawMux, RawModes);
//TrcPrint("COOKED: countervalue: %x cbox1 = %x cbox2 = %x\n", CounterValue, cbox1, cbox2);
    AxpWritePerformanceCounter( CounterValue,
                                (ULONG) 0,
                                cbox1,
                                cbox2);



    return STATUS_SUCCESS;
}

VOID
DisablePerformanceCounter(
    PDEVICE_EXTENSION DeviceExtension
    )
/*++

Routine Description:

    Disable the performance counter from counting events.

    N.B. - Expected to be synchronized with respect to the interrupt
           handler and any other writers of the device extension.

Arguments:

    DeviceExtension - Supplies a pointer to the device extension for the
                          performance counter to be disabled.

Return Value:

    None.

--*/
{
//	if( DeviceExtension->Started == TRUE )
//	  {
		AxpWritePerformanceCounter((ULONG)0,
		  (ULONG)0,
		  (ULONG)0,
		  (ULONG)0 );
//	  }

	return;
}

unsigned long int DeriveShiftValue(unsigned long int mask)
  {
	unsigned long int ShiftValue = 0;

	while ((mask & 0xf) == 0)
	  {
		ShiftValue += 4;
		mask >>= 4;
	  }

	while((mask & 0x1) == 0)
	  {
		ShiftValue++;
		mask >>= 1;
	  }

	return ShiftValue;
  }