/*++

Copyright (c) 1993-1994  Digital Equipment Corporation

Module Name:

    nta21064.c

Abstract:

    This module implements the routines that interface directly to the
    Alpha AXP 21064 performance counter driver via the NT I/O APIs.

Author:

    Joe Notarangelo  25-Jun-1993

Revision History:

    Joe Notarangelo  05-Feb-1994
        Added support for the external counter selects for the 21066.

--*/

//#include "nt.h"
#include <ntddk.h>

//
// Include the interface defintions.
//

#include "stdio.h"
#include "nta21064.h"
//#include "ntrtl.h"

typedef ULONGLONG AXPCOUNTER;

NTSTATUS
OpenAxp21064Counter(
    AXP21064_PCCOUNTER Counter,
    HANDLE *FileHandle
    )
/*++

Routine Description:

    This routine opens the performance counter device and returns both
    the status and the file handle associated with the open.

Arguments:

    Counter - Supplies the number of the counter to open.

    FileHandle - Receives the file handle returned from NtOpen

Return Value:

    If the Counter is valid, the status from NtOpen is returned, otherwise
    STATUS_INVALID_PARAMETER is returned.


    N.B - This interface is obsolete, OpenMpAxp21064Counter is the
          preferred and complete API.  This function is preserved for
          backward compatibility and can only interface with processor 0
          counters.

--*/
{
    UCHAR DeviceNameBuffer[80];
    IO_STATUS_BLOCK IoStatus;
    OBJECT_ATTRIBUTES ObjectAttributes;
    NTSTATUS Status;
    UNICODE_STRING UnicodeDeviceName;


    switch( Counter ){

    case PerformanceCounter0:
    case PerformanceCounter1:

        break;

    default:

        //
        // Invalid performance counter.
        //

        return STATUS_INVALID_PARAMETER;

    }

    sprintf( DeviceNameBuffer, "%sP%dPC%d",
                DD_AXP21064_PERFORMANCE_DEVICE_NAME,
                PROCESSOR_0,
                Counter );

    RtlCreateUnicodeStringFromAsciiz( &UnicodeDeviceName, DeviceNameBuffer );

    InitializeObjectAttributes( &ObjectAttributes,
                                &UnicodeDeviceName,
                                0,
                                NULL,
                                NULL );

    Status = NtCreateFile( FileHandle,
                           FILE_READ_DATA | FILE_WRITE_DATA,
                           &ObjectAttributes,
                           &IoStatus,
                           (PLARGE_INTEGER)NULL,
                           0L,
                           0L,
                           FILE_OPEN_IF,
                           0L,
                           (PVOID)NULL,
                           0L );

    if( Status == STATUS_SUCCESS ){
        Status = IoStatus.Status;
    }

    RtlFreeUnicodeString( &UnicodeDeviceName );

    return Status;

}

NTSTATUS
OpenMpAxp21064Counter(
    ULONG ProcessorNumber,
    AXP21064_PCCOUNTER Counter,
    HANDLE *FileHandle
    )
/*++

Routine Description:

    This routine opens the performance counter device and returns both
    the status and the file handle associated with the open.

Arguments:

    ProcessorNumber - Supplies the number of the processor of the counter to
                        open.

    Counter - Supplies the number of the counter to open.

    FileHandle - Receives the file handle returned from NtOpen

Return Value:

    If the Counter is valid, the status from NtOpen is returned, otherwise
    STATUS_INVALID_PARAMETER is returned.

--*/
{
    UCHAR DeviceNameBuffer[80];
    IO_STATUS_BLOCK IoStatus;
    OBJECT_ATTRIBUTES ObjectAttributes;
    NTSTATUS Status;
    UNICODE_STRING UnicodeDeviceName;


    if( ProcessorNumber >= MAXIMUM_PROCESSORS ){
        return STATUS_INVALID_PARAMETER;
    }

    switch( Counter ){

    case PerformanceCounter0:
    case PerformanceCounter1:

        break;

    default:

        //
        // Invalid performance counter.
        //

        return STATUS_INVALID_PARAMETER;

    }

    sprintf( DeviceNameBuffer, "%sP%dPC%d",
                DD_AXP21064_PERFORMANCE_DEVICE_NAME,
                ProcessorNumber,
                Counter );

    RtlCreateUnicodeStringFromAsciiz( &UnicodeDeviceName, DeviceNameBuffer );

    InitializeObjectAttributes( &ObjectAttributes,
                                &UnicodeDeviceName,
                                0,
                                NULL,
                                NULL );

    Status = NtCreateFile( FileHandle,
                           FILE_READ_DATA | FILE_WRITE_DATA,
                           &ObjectAttributes,
                           &IoStatus,
                           (PLARGE_INTEGER)NULL,
                           0L,
                           0L,
                           FILE_OPEN_IF,
                           0L,
                           (PVOID)NULL,
                           0L );

    if( Status == STATUS_SUCCESS ){
        Status = IoStatus.Status;
    }

    RtlFreeUnicodeString( &UnicodeDeviceName );

    return Status;

}

NTSTATUS
CloseAxp21064Counter(
    HANDLE FileHandle
    )
/*++

Routine Description:

    This routine closes a performance counter device.

Arguments:

    FileHandle - Supplies the handle to the open performance counter device.

Return Value:

    The return status from NtClose is returned.

--*/
{

    return( NtClose(FileHandle) );

}

NTSTATUS
ReadAxp21064Counter(
    HANDLE FileHandle,
    AXPCOUNTER *CounterValue
    )
/*++

Routine Description:

    Read the current value of the performance counter.

Arguments:

    FileHandle - Supplies the file handle of the performance counter device.

    CounterValue - Receives the counter value returned by NtReadFile.

Return Value:

    The return status from NtReadFile is returned.

--*/
{
    LARGE_INTEGER ByteOffset;
    IO_STATUS_BLOCK IoStatus;
    NTSTATUS Status;

    ByteOffset.QuadPart = 0;

    Status = NtReadFile( FileHandle,
                         (HANDLE)NULL,
                         (PIO_APC_ROUTINE)NULL,
                         NULL,
                         &IoStatus,
                         CounterValue,
                         sizeof(AXPCOUNTER),
                         &ByteOffset,
                         (PULONG)NULL );

    if( Status == STATUS_SUCCESS ){
        Status = IoStatus.Status;
    }

    return Status;

}


NTSTATUS
ReadAxp21064PalCounters(
    HANDLE FileHandle,
    AXP21064_RDPALCOUNTERS *PalCounters
    )
/*++

Routine Description:

    Read the PALcode event counters.

Arguments:

    FileHandle - Supplies the file handle of the performance counter device.

    PalCounters - Supplies a pointer to the read PAL counters structure to
                  receive the read data.

Return Value:

    The status of the device io control system service is returned.
    If the returned size is not expected then STATUS_UNSUCCESSFUL will be
    returned if the other statuses are otherwise STATUS_SUCCESS.

--*/
{
    IO_STATUS_BLOCK IoStatus;
    NTSTATUS Status;

    Status = NtDeviceIoControlFile( FileHandle,
                                    (HANDLE)NULL,
                                    (PIO_APC_ROUTINE)NULL,
                                    (PVOID)NULL,
                                    &IoStatus,
                                    (unsigned long)IOCTL_AXP21064_RDPALCOUNTERS,
                                    NULL,
                                    0,
                                    PalCounters,
                                    sizeof(AXP21064_RDPALCOUNTERS) );

    if( Status == STATUS_SUCCESS ){
        Status = IoStatus.Status;
        if( (Status == STATUS_SUCCESS) &&
            (IoStatus.Information != sizeof(AXP21064_RDPALCOUNTERS)) ){
            Status = STATUS_UNSUCCESSFUL;
        }
    }

    return Status;
}

NTSTATUS
SetAxp21064CounterValue(
    HANDLE FileHandle,
    AXPCOUNTER CounterValue
    )
/*++

Routine Description:

    Write a new value to the performance counter.

Arguments:

    FileHandle - Supplies the file handle for the performance counter device.

    CounterValue - Supplies the new value to write to the performance counter.

Return Value:

    The return status from NtWriteFile is returned.

--*/
{
    LARGE_INTEGER ByteOffset;
    AXPCOUNTER CounterBuffer = CounterValue;
    IO_STATUS_BLOCK IoStatus;
    NTSTATUS Status;

    ByteOffset.QuadPart = 0;

    Status = NtWriteFile( FileHandle,
                          (HANDLE)NULL,
                          (PIO_APC_ROUTINE)NULL,
                          NULL,
                          &IoStatus,
                          &CounterBuffer,
                          sizeof(AXPCOUNTER),
                          &ByteOffset,
                          (PULONG)NULL );

    if( Status == STATUS_SUCCESS ){
        Status = IoStatus.Status;
    }

    return Status;

}

NTSTATUS
ResetAxp21064CounterValue(
    HANDLE FileHandle
    )
/*++

Routine Description:

    Reset the performance counter value to zero.

Arguments:

    FileHandle - Supplies the file handle for the performance counter device.

Return Value:

    The return value from NtWriteFile is returned.

--*/
{

    return( SetAxp21064CounterValue( FileHandle, 0 ) );

}

NTSTATUS
StartAxp21064Counter(
    HANDLE FileHandle
    )
/*++

Routine Description:

    Start the performance counter vi the ioctl driver interface.

Arguments:

    FileHandle - Supplies the file handle for the performance counter device.

Return Value:

    The value returned by NtDeviceIoControlFile.

--*/
{
    IO_STATUS_BLOCK IoStatus;
    NTSTATUS Status;

    Status = NtDeviceIoControlFile( FileHandle,
                                    (HANDLE)NULL,
                                    (PIO_APC_ROUTINE)NULL,
                                    NULL,
                                    &IoStatus,
                                    (unsigned long)IOCTL_AXP21064_STARTCOUNTER,
                                    NULL,
                                    0L,
                                    NULL,
                                    0L );

    if( Status == STATUS_SUCCESS ){
        Status = IoStatus.Status;
    }

    return Status;

}

NTSTATUS
StopAxp21064Counter(
    HANDLE FileHandle
    )
/*++

Routine Description:

    Stop the performance counter via the ioctl driver interface.

Arguments:

    FileHandle - Supplies the file handle for the performance counter device.

Return Value:

    The value returned by NtDeviceIoControlFile.

--*/
{
    IO_STATUS_BLOCK IoStatus;
    NTSTATUS Status;

    Status = NtDeviceIoControlFile( FileHandle,
                                    (HANDLE)NULL,
                                    (PIO_APC_ROUTINE)NULL,
                                    NULL,
                                    &IoStatus,
                                    (unsigned long)IOCTL_AXP21064_STOPCOUNTER,
                                    NULL,
                                    0L,
                                    NULL,
                                    0L );

    if( Status == STATUS_SUCCESS ){
        Status = IoStatus.Status;
    }

    return Status;
}

NTSTATUS
MuxControlAxp21064Counter(
    HANDLE FileHandle,
    AXP21064_PCMUXCONTROL MuxControl
    )
/*++

Routine Description:

    Set the mux control of the performance coutner via the ioctl device
    interface.

Arguments:

    FileHandle - Supplies the file handle for the performance counter device.

    MuxControl - Supplies the new mux control to set the performance counter.

Return Value:

    The status returned by NtDeviceIoControlFile.

--*/
{
    IO_STATUS_BLOCK IoStatus;
    NTSTATUS Status;
    AXP21064_SETMUXCONTROL SetMuxControl;

    SetMuxControl.MuxControl = MuxControl;

    Status = NtDeviceIoControlFile( FileHandle,
                                    (HANDLE)NULL,
                                    (PIO_APC_ROUTINE)NULL,
                                    NULL,
                                    &IoStatus,
                                    (unsigned long)IOCTL_AXP21064_SETMUXCONTROL,
                                    &SetMuxControl,
                                    sizeof(AXP21064_SETMUXCONTROL),
                                    NULL,
                                    0 );

    if( Status == STATUS_SUCCESS ){
        Status = IoStatus.Status;
    }

    return Status;

}

NTSTATUS
EventCountAxp21064Counter(
    HANDLE FileHandle,
    AXP21064_PCEVENTCOUNT EventCount
    )
/*++

Routine Description:

    Set the event count of the performance counter via the ioctl device
    interface.

Arguments:

    FileHandle - Supplies the file handle for the performance counter device.

    EventCount - Supplies the new event count to set the performance counter.

Return Value:

    The status returned by NtDeviceIoControlFile.

--*/
{
    IO_STATUS_BLOCK IoStatus;
    NTSTATUS Status;
    AXP21064_SETCOUNTER SetCounter;

    SetCounter.EventCount = EventCount;

    Status = NtDeviceIoControlFile( FileHandle,
                                    (HANDLE)NULL,
                                    (PIO_APC_ROUTINE)NULL,
                                    NULL,
                                    &IoStatus,
                                    (unsigned long)IOCTL_AXP21064_SETCOUNTER,
                                    &SetCounter,
                                    sizeof(AXP21064_SETCOUNTER),
                                    NULL,
                                    0 );

    if( Status == STATUS_SUCCESS ){
        Status = IoStatus.Status;
    }

    return Status;

}

NTSTATUS
SelectAxp21066ExternalCounter(
    HANDLE FileHandle,
    AXP21066_EXTERNALSELECT ExternalSelect
    )
/*++

Routine Description:

    Select the type of event to count in the 21066 external counter
    via the ioctl device interface.

    N.B. - This routine is only useful on systems that utilize a
           21066 processor.

Arguments:

    FileHandle - Supplies the file handle for the performance counter device.

    ExternalSelect - Supplies the new selection for the external counter.

Return Value:

    The status returned by NtDeviceIoControlFile.

--*/
{
    IO_STATUS_BLOCK IoStatus;
    NTSTATUS Status;
    AXP21066_EXTERNALSELECTION ExternalSelection;

    ExternalSelection.ExternalSelect = ExternalSelect;

    Status = NtDeviceIoControlFile( FileHandle,
                                    (HANDLE)NULL,
                                    (PIO_APC_ROUTINE)NULL,
                                    NULL,
                                    &IoStatus,
                                    (unsigned long)IOCTL_AXP21066_EXTERNALSELECT,
                                    &ExternalSelection,
                                    sizeof(AXP21066_EXTERNALSELECTION),
                                    NULL,
                                    0 );

    if( Status == STATUS_SUCCESS ){
        Status = IoStatus.Status;
    }

    return Status;

}


