/****************************************************************************/
/*                                                                          */
/*  COPYRIGHT (c) 1993-1999 BY                                              */
/*  Digital Equiptment/Compaq Computer Corporation.                         */
/*  ALL RIGHTS RESERVED.                                                    */
/*                                                                          */
/*  This file is part of the Iprobe performance suite for Alpha processors. */
/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307    */
/*                                                                          */
/****************************************************************************/
/*
*
* Program: REP.c  (Read Entry Points) 
*
* Abstract: Module containing all the code to create the address map of a user
* program, including any user and kernel processes. 
*
* Author: Satyamahesh Avala 
* 
* Creation Date: 06-May-94
*
* Modification History:
*
* Phil Ezolt: 8-Dec-98
*     Problem:  Rep had much extra Digital Unix cruft left in.  
*     Solution: Removed extra stuff. 
*
* Phil Ezolt: 8-Dec-98
*     Problem: Rep can't analyze by pid or look at dynamic address ranges. 
*     Solution: Add support for analysis by pid using proc.   This does not 
*               currently work unless the program is already running.  That 
*               will be fixed. 
*
* Phil Ezolt: 23-Sept-98
*     Problem: Rep can only process kernel symbols
*     Solution: Add support using bfd for extracting symbols from static images. 
*
* Phil Ezolt: 15-Sept-98:  This is the first port of REP to linux.
*     Problem: Previously only digital unix version.
*     Solution: We can currently grab kernel symbols from the kernel map that it keeps around after compiling.
*
* Phil Ezolt: 15-Jun-97 
*     Problem:  Rep would wait for the image that it was going to analyze to exit before it would 
*               start the analysis.  This would appear to be a "hang" in rep when it was processing a
*               program that didn't execute immediately. 
*     Solution: Removed the code the code that waited for the child process to exit.  However, this caused
*               the child process to be analyzed before it had execed the required image.  Solved THIS
*               by waiting until the name of the child process was something other than the called 
*               program name.  Now user can't run rep on the version of rep they are running.  Not a big deal.
*
* Phil Ezolt: 25-Jun-97 
*     Problem:  Global Variables, Out of date comments.
*     Solution:  Removed global variables, by putting them in wrapper functions, and updated comments.
*
* Phil Ezolt: 18-Jun-97 
*     Problem:  Not enough common code, some cryptic.
*     Solution: Removed the need for the external "find_file" routine. 
*               Finished unified command line args parsing using getopts.
*               Finished unifing the Appending of input files. 
*               Change Symbol Manipulation for the unsupported "obj" functions to
*               the support st_ functions.
*               Removed most global variables. 
*               Added Proper Prototypes.
*
* Phil Ezolt: 17-Jun-97 
*     Problem:  Unecessary Bloat
*     Solution: Removed PME Bit support, and started to unify command line parsing. 
*
* Phil Ezolt: 3-Jun-97 
*     Problem:  Code hard to read and functions not very well commented. 
*     Solution: Renamed varibles and added block comments before functions, moved 
*               some constants to  #defines.
*
* Satyamahesh Avala: 30-Nov-94
*     Problem: 
*     Solution: Included another source file, repfind.c, which has the necessary routines
*               to find the file name on disk, given a process_id number. 'rep' is
*               interfaced with 'find_file' routine.
*               A bug which kills the running process if 'rep' not able to attach to that
*               process is cleaned.
* 
* Satyamahesh Avala: 17-Sep-94
*     Problem: 
*     Solution: Implemented to get the file name from process structure, and deleted 
*               the module 'get_filename'.
*               Included checks to see if object file is stripped.
*               Included 'StartStaticImage' module to kick off static image if PME bit set.
* 
* Satyamahesh Avala: 16-Sep-94
*    Problem: 
*    Solution: Modified to take care of static and dynamic executables. 
*              Routines which look for file headers and symbol tables are re-written
*              using the new interface(obj.h).
* 
* Satyamahesh Avala: 10-Jul-94
*    Problem: 
*    Solution: Added setting PME bit for any user specified process.
* 
* Satyamahesh Avala: 17-Jun-94
*    Problem: 
*    Solution: Modified REP to merge RESOLVE_ADDRESSES and to make a single
*              image of REP and RESOLVE_ADDRESSES.
* 
*/

#define MAX_PATHNAME_SIZE 250
#define STRINGLENGTH 250    



#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

/* This is the GNU way of manipulating binaries */
#ifdef __FreeBSD__
#include <sys/sysctl.h>
#define FREEBSD_BFD
#endif
#include <bfd.h>

#include "stdtypes.h"
#include "rep_common.h"


/************************************************************************************/
/* Name: SaveAllSymbolsFromObject                                                   */
/*                                                                                  */
/* Purpose: Given a pointer to the bfd of a file, this will grab all of the         */
/*       Functions within a file, and store them in our function/address list       */
/*                                                                                  */
/* Arguments:  CommandFlags: A structure describing the command line args.          */
/*             bfd: A descriptor that we can use to extract information about the   */
/*                  file.                                                           */
/*             StartingOffset:  Memory location where to current object is loaded to*/
/*                                                                                  */
/* Notes:  None.                                                                    */
/*                                                                                  */
/* Outputs: None                                                                    */
/*                                                                                  */
/* Function Value:  TRUE is returned if it was successful, FALSE if it failed.      */
/*                                                                                  */
/************************************************************************************/

boolean GetSymbolsFromBfd(CMD_TABLE *CommandFlags, bfd *abfd, uint64 StartingOffset)
{
  uint64 storage_needed,NumberOfSymbols,EndAddress;

  asymbol **SymbolTable;
  char *FileName;
  int CurrentSymbolNumber;
  asection *section;
  uint64 TextStart;
  

  /* This check to see if we are looking at a dynamically loaded library. */
  /* If the object is not dynamically loaded into memory, the offset we were given
     is meaningless.  Ignore it.  The executable will tell us where it is to be loaded. */

  if (!(bfd_get_file_flags(abfd) & DYNAMIC))
    {
      StartingOffset=0;
    }
  

  /* Every symbol that this inserts spans to the end of the address range.  I think that this
     could actually be done right.  We have the beginning and ending addresses, so this
     should probably be changed. 
  */
  
  EndAddress=0xffffffffffffffff;
 
  FileName=bfd_get_filename(abfd);

  /* Add a dummy entry that spans from the entry point of the image to the end of the address space. */
  /* Right now, if we only have a static image, it will not be start at the
     proper location, instead it will start at zero.  I'll have to work on
     this. --Phil
  */
  
  //  AddSymbolToList("Unknown",FileName, StartingOffset, EndAddress);


  /* This occurs if the symbol tables are stripped */

  if (!(bfd_get_file_flags(abfd) & HAS_SYMS))
    {
      printf("There are no symbols in %s\n",FileName);
      return;
    }


  storage_needed = bfd_get_symtab_upper_bound (abfd);

  if (storage_needed < 0)  
    {
      printf("I can't figure out how much storage you need for your users symbols. Error.");
      return;
    } 

  
  if (storage_needed == 0) {    
    printf("There are no symbols in %s\n",FileName);
    return ;    
  }        

  SymbolTable = (asymbol **) malloc (storage_needed); 
  NumberOfSymbols = bfd_canonicalize_symtab (abfd, SymbolTable); 

  if (NumberOfSymbols < 0)   
      {
	printf("I can't figure out the number of users symbols. Error.");
	free(SymbolTable);
	return;	
      }

  if (NumberOfSymbols == 0)   
      {
	printf("There are no symbols in %s\n",FileName);
	free(SymbolTable);
	return;	
      }

  section = bfd_get_section_by_name(abfd, ".text");
  TextStart = bfd_get_section_vma (abfd, section);
  for (CurrentSymbolNumber = 0; CurrentSymbolNumber < NumberOfSymbols; CurrentSymbolNumber++) {   
      /*
       * Only add it if it is a function. 
       * If this fails, also check if its in the text segement, since
       * the BSF_FUNCTION flags is limited to elf, and we occasionally 
       * run osf/1 ecoff binaries under emulation
       */
      if(((SymbolTable[CurrentSymbolNumber]->flags)&BSF_FUNCTION) 
	      || ((SymbolTable[CurrentSymbolNumber]->section->vma) == TextStart))
	{	  
	  AddSymbolToList(SymbolTable[CurrentSymbolNumber]->name,FileName, 
			  (StartingOffset+SymbolTable[CurrentSymbolNumber]->section->vma + SymbolTable[CurrentSymbolNumber]->value), 
			  EndAddress);
	}  
      
    }    
    free(SymbolTable);
    return TRUE;
      
}


/************************************************************************************/
/* Name: GetSymbolsFromObject                                                       */
/*                                                                                  */
/* Purpose:   Given an object this will extract all of the symbols.                 */
/*                                                                                  */
/* Arguments:  CommandFlags: A structure describing the command line args.          */
/*             ProcessName: This will eventually tell us how much off of            */
/*             StartingOffset: This will eventually tell us how much off of         */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/* Outputs: None.                                                                   */
/*                                                                                  */
/* Function Value: None                                                             */
/*                                                                                  */
/************************************************************************************/

static void GetSymbolsFromObject(CMD_TABLE *CommandFlags, char *ProcessName, uint64 StartingOffset)
{
  bfd *abfd;
  char **matching;

  printf("\tExtracting symbols from:%s\n",ProcessName);

  /* Open the file we want to analyze */

  bfd_init();
  abfd=bfd_openr(ProcessName,NULL);
  
  if (abfd == NULL)
    {
      fprintf(stderr, "Error: can't open the image %s for object analysis.\n"
	      "The specified file may be a shell script or batch file.\n", 
	      ProcessName);      
      exit(EXIT_FAILURE);
    }


  /* Although this check is completely (I think) not required by the API of bfd, 
     If I remove it, the program will core dump.   */
  
  if (!bfd_check_format_matches (abfd, bfd_object, &matching))
    {
      printf("This is not an object file.\n");
    }
  
  GetSymbolsFromBfd(CommandFlags, abfd, StartingOffset);
  
  bfd_close(abfd);
  
}

/************************************************************************************/
/* Name: GetTextStartofObject                                                       */
/*                                                                                  */
/* Purpose:   Given an object this will extract the location of its text segment    */
/*                                                                                  */
/* Arguments: ImageName: Name of executable whose text segment we'd  like to find   */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/* Outputs: None.                                                                   */
/*                                                                                  */
/* Function Value: None                                                             */
/*                                                                                  */
/************************************************************************************/

static uint64 GetTextFromObject(const char *ImageName)
{
  bfd *abfd;
  char **matching;
  asection *section;
  uint64 TextStart;
  printf("Extracting symbols from:%s\n",ImageName);

  /* Open the file we want to analyze */

  bfd_init();
  abfd=bfd_openr(ImageName,NULL);
  
  if (abfd == NULL)
    {
      fprintf(stderr, "Error: can't open the image %s for object analysis.\n"
	      "The specified file may be a shell script or batch file.\n", 
	      ImageName);      
      exit(EXIT_FAILURE);
    }


  /* Although this check is completely (I think) not required by the API of bfd, 
     If I remove it, the program will core dump.   */
  
  if (!bfd_check_format_matches (abfd, bfd_object, &matching))
    {
      printf("This not is an object file.\n");
    }
  
  section = bfd_get_section_by_name(abfd, ".text");
  TextStart = bfd_get_section_vma (abfd, section);
/*  printf("%s's text segment starts at 0x%lx\n", ImageName, TextStart);*/
  bfd_close(abfd);
  return(TextStart);
}



/************************************************************************************/
/* Name: FindObjectsByCmd                                                           */
/*                                                                                  */
/* Purpose:  Given a filename, this will save all of the process's address space info. */
/*                                                                                  */
/* Arguments:  CommandFlags: A structure describing the command line args.          */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/* Outputs: None.                                                                   */
/*                                                                                  */
/* Function Value: None                                                             */
/*                                                                                  */
/************************************************************************************/

void FindObjectsByCmd(CMD_TABLE *CommandFlags, char *CmdFileName, char *CmdName)
{
    FILE *MapFile;
    char MapFileName[MAX_PATHNAME_SIZE];
    char c;

    uint64 StartingAddress;
    char ObjectName[MAX_PATHNAME_SIZE];
  
 
    /* 
     * This is really easy on FreeBSD.  We first use objdump to find
     * the text segment, then we look at the output of ldd, which should
     * look something like:
     *    ldd ./iprobe_test
     *    ./iprobe_test:
     *        libc.so.3 => /usr/lib/libc.so.3 (0x160108000)
     */

    StartingAddress = GetTextFromObject(CmdFileName);
    GetSymbolsFromObject(CommandFlags,CmdFileName, StartingAddress);
/* if its dynamic, then map its' libraries */
    
    sprintf(MapFileName, "/usr/bin/ldd %s\n", CmdFileName);
    if ((MapFile=popen(MapFileName,"r"))==NULL)
	    return;
    for(c = ' '; c!='\n' && c!=EOF; c = fgetc(MapFile))
	    ;
    while(!feof(MapFile)){
	    fscanf(MapFile, "%*s %*s %s %*c %lx %*c", &ObjectName, &StartingAddress);
	    GetSymbolsFromObject(CommandFlags,ObjectName, StartingAddress);
    }
    pclose(MapFile);
}

/************************************************************************************/
/* Name: FindObjectsByPid                                                           */
/*                                                                                  */
/* Purpose:  Given a PID this will save all of the process's address space info.    */
/*                                                                                  */
/* Arguments:  CommandFlags: A structure describing the command line args.          */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/* Outputs: None.                                                                   */
/*                                                                                  */
/* Function Value: None                                                             */
/*                                                                                  */
/************************************************************************************/

void FindObjectsByPid(CMD_TABLE *CommandFlags, uint64 pid)
{
  FILE *CmdFile;
  char CmdName[MAX_PATHNAME_SIZE], CmdFileName[MAX_PATHNAME_SIZE];

  /*
   *  This is the name of the FreeBSD proc file that points at the
   *  running executable 
   */

  sprintf(CmdFileName,"/proc/%d/file",pid);
  sprintf(CmdName,"/proc/%d/cmdline",pid);
  if ((CmdFile=fopen(CmdName,"r"))==NULL)
    {
      printf("Can't open the command name file of process %ld. \n", pid);
      return;
    }
  if (fscanf(CmdFile,"%s\n", CmdName) <=0)
    {
      printf("Can't read the command name file of process %ld. \n", pid);
      return;
    }
  fclose(CmdFile);
  FindObjectsByCmd(CommandFlags, CmdFileName, CmdName);
}


/************************************************************************************/
/* Name:  ExtractUserSymbols                                                        */
/*                                                                                  */
/* Purpose: This function does the following things:                                */
/*          1) Fork off a the program we want to analyze.                           */
/*          2) Wait until it is running.                                            */
/*          3) Analyze it by pid.                                                   */
/*          4) When we're done, kill it.                                            */
/*                                                                                  */
/* Arguments: argv: This contains a modified argv that will be passed into the      */
/*                     child process.                                               */
/*            CommandFlags: A structure describing the command line args.           */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/* Outputs: None                                                                    */
/*                                                                                  */
/* Function Value: None                                                             */
/*                                                                                  */
/************************************************************************************/

void ExtractUserSymbols(char *argv[], CMD_TABLE *CommandFlags)
{
  FindObjectsByCmd(CommandFlags, argv[0], argv[0]);
}





/************************************************************************************/
/* Name: ExtractKernelSymbols                                                       */
/*                                                                                  */
/* Purpose: This will extract all of the symbols from the kernel.                   */
/*                                                                                  */
/* Arguments: CommandFlags: A structure describing the command line args.           */
/*                                                                                  */
/* Notes:  Looks at the sysctl mib KERN_BOOTFILE to find the booted kernel          */
/*                                                                                  */
/* Outputs:  None                                                                   */
/*                                                                                  */
/* Function Value: None                                                             */
/*                                                                                  */
/************************************************************************************/
#define MOD_PATH_LEN       10
#define MOD_PATH_ITEM_LEN 100

static void ExtractKernelSymbols(CMD_TABLE *CommandFlags)
{
  char c, *cptr;
  char bootfile[100], ModListName[100],ObjectName[100],ModuleName[100];
  char *ModPath[MOD_PATH_LEN];
  char ModPathString[MOD_PATH_LEN*MOD_PATH_ITEM_LEN+MOD_PATH_LEN];
  uint64 StartingAddress;
  FILE *KernelSymbolFile,*ModList;
  int i, Id, ModPathLen = 0;
  char **ap, *inputstring;
  size_t len = 100;
  char Path[100];
  struct stat sb;

  sysctlbyname("kern.bootfile", bootfile, &len, NULL, 0);
  len = sizeof(ModPathString);
  sysctlbyname("kern.module_path", ModPathString, &len, NULL, 0);

/* should return something like '/;/boot/;/modules/' */
  cptr = &ModPathString[0];
  for (ap = ModPath; (*ap = strsep(&cptr, ";")) !=  NULL;){
	  if (**ap != '\0'){
		  ModPathLen++;
		  if (++ap >= &ModPath[MOD_PATH_LEN])
			  break;
	  }
  }

  printf("Reading kernel symbols:\n");

  sprintf(ModListName, "/sbin/kldstat\n");
  if ((ModList=popen(ModListName,"r"))==NULL)
	  return;
  for(c = ' '; c!='\n' && c!=EOF; c = fgetc(ModList))
	  ;
  
   while(!feof(ModList)){
	   StartingAddress =0; ObjectName[0]=0;
	   fscanf(ModList, "%d %*d %lx %*s %s\n",
		  &Id, &StartingAddress, &ObjectName);
	   if(Id == 1) {
		   /* 
		    *  this is the kernel.  Use the load address we
		    *  found here & the bootfile path we got via sysctl
		    *  I *think* that the module name may be wrong when 
		    *  kern.bootfile != /kernel, hence the special case 
		    */
		   GetSymbolsFromObject(CommandFlags,bootfile, StartingAddress);
		   continue;
	   }
	   for(i = 0; i < ModPathLen; i++){
		   strcpy(Path, ModPath[i]);
		   strcat(Path, ObjectName);
		   if(!stat(Path, &sb)){
			   GetSymbolsFromObject(CommandFlags,Path, StartingAddress);
		   }
	   }

    }
    pclose(ModList);
  return;
}

/************************************************************************************/
/* Name: main()                                                                     */
/*                                                                                  */
/* Purpose:  1) First, analyze the different command line switches, and set flags   */
/*              according to the passed in options.                                 */
/*           2) Given the preceding flags, preform specific actions.                */
/*           3) Close output file, and possibly resolve the output file.            */
/*                                                                                  */
/*                                                                                  */
/* Arguments: argc: Number of arguments.                                            */
/*            argv: Vector of the different arguments (0 being the name of the      */
/*            program.                                                              */
/*                                                                                  */
/* Notes: None.                                                                     */
/*                                                                                  */
/* Outputs: None                                                                    */
/*                                                                                  */
/* Function Value: The error code passed to the operating systems.                  */
/*                                                                                  */
/************************************************************************************/


int main (int argc, char *argv[]) 
{
  CMD_TABLE *CommandFlags;
  PFILELIST pInputFiles;  
  
  CommandFlags = GetArgsAndSwitches(argc, argv);
  
  if (!CommandFlags->Valid)
  {
    printf("Invalid command line arguments use -h for help\n");
    exit(EXIT_FAILURE);
  }

  if (CommandFlags->Help)
  {
    SendHelp();
    exit(EXIT_SUCCESS);
  }

  if (!OpenRepFile(CommandFlags->OutputFileName, CommandFlags))
  {
    fprintf(stderr, "Error: Output file can't be created\n");
    perror("fopen");
    exit(EXIT_FAILURE);
  }

  /* Read input file. */
  
  if (pInputFiles = CommandFlags->InputFileNames)
  {
    while (pInputFiles != NULL)
    {
      if (!AppendRepFile(pInputFiles->FileName, CommandFlags))
      {
				break;
      }

      pInputFiles = (PFILELIST) pInputFiles->Next;
    }
  }
  
  fprintf(stdout, "Your output file will be  %s\n", CommandFlags->OutputFileName);

  if (CommandFlags->FResolve == FALSE) 
  {
    if (CommandFlags->UMode == TRUE)
    {
      if (CommandFlags->ArgStr[0] != NUL)
	{
	  
	  if (CommandFlags->Logging)
	    {
	      fprintf(stderr, "Analyzing Image.\n");	  
	    }
	  
	  ExtractUserSymbols(&argv[CommandFlags->FirstNonFlagArgumentNumber], CommandFlags);
	}		  
      else
	{
	  if (CommandFlags->ProcId != 0)
	    {
	      if (CommandFlags->Logging)
		{
		  fprintf(stderr,"Analyzing Running Process. Pid: %d\n", CommandFlags->ProcId);	  
		}	    
	      FindObjectsByPid(CommandFlags,CommandFlags->ProcId);
	    }	
	}
    }
    
    
    if (CommandFlags->KMode == TRUE)
      {
	ExtractKernelSymbols(CommandFlags);
      }
  }
  
  CloseRepFile();
  
  if (OrderSymbolList()) {
    if (!CommandFlags->NoResolve) 
      {
	ResolveAddress(CommandFlags->Logging, CommandFlags->OutputFileName);
      }
    else
      {
	if (OpenRepFile(CommandFlags->OutputFileName, CommandFlags))
	  {
	    WriteRep(CommandFlags);
	  }
	else
	  {
	    fprintf(stderr, "Error: Output file can't be created\n");
	    perror("fopen");
	  }  
      }
  }
  /* Free up the memory for the Structure containing the all of the options */
  free(CommandFlags);
}

